# 786
# Aldy source: gene.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import Tuple, Dict, List, Optional, Set, NamedTuple
from dataclasses import dataclass, field
from enum import Enum
from natsort import natsorted
import os
import yaml
import collections

from .common import (
    GRange,
    AldyException,
    allele_name,
    sorted_tuple,
    seq_to_amino,
    rev_comp,
    log,
    freezekey,
)


class Mutation(NamedTuple):
    """
    Mutation description. Immutable.

    :param pos: Reference genome position (0-based).
    :param op: Variant operation in HGVS-like format:
    - `X>Y`:  a SNP from X to Y
    - `insX`:  an insertion of X
    - `delX`:  a deletion of X (X, Y are both of format `[ACGT]+`).
    """

    pos: int
    op: str

    def __str__(self):
        return f"{self.pos + 1}.{self.op}"


@dataclass
class MinorAllele:
    """
    Minor allele description.

    :param name: Minor allale name.
    :param alt_name: List of alternative names.
    :param neutral_muts: Netural mutations that describe the minor allele.
    :param activity: Activity indicator (see PharmVar for details).
    :param evidence: Evidence indicator (see PharmVar for details).
    :param pharmvar: PharmVar ID.
    """

    name: str
    alt_name: Optional[str] = None
    neutral_muts: Set[Mutation] = field(default_factory=set)
    _activity: Optional[str] = "unknown"
    evidence: Optional[str] = None
    pharmvar: Optional[str] = None

    def __str__(self):
        muts = ", ".join(map(str, self.neutral_muts))
        return f"Minor({self.name}; [{muts}])"

    @property
    def activity(self):
        return self._activity or "unknown"


@dataclass
class MajorAllele:
    """
    Major allele description.

    :param name: Major allele name.
    :param cn_config: Copy-number configuration.
    :param func_muts: Functional mutations that describe the major allele.
    :param minors: Minor alleles (suballeles) that are derived from this major allele.
    """

    name: str
    cn_config: str = "1"
    func_muts: Set[Mutation] = field(default_factory=set)
    minors: Dict[str, MinorAllele] = field(default_factory=dict)

    def get_minor_mutations(self, minor: str):
        """Sequence of mutations that describe a minor allele."""
        for m in self.func_muts:
            yield m
        for m in self.minors[minor].neutral_muts:
            yield m


class CNConfigType(Enum):
    """Enumeration describing the type of a copy-number configuration"""

    DEFAULT = 0
    LEFT_FUSION = 1
    RIGHT_FUSION = 2
    DELETION = 3
    CUSTOM = 4


@dataclass
class CNConfig:
    """
    Copy-number (CN) configuration description. Immutable.

    :param cn:
        Value of the expected region copy number in each gene.
        For example, `cn[0]["e1"] == 1` means that the exon 1 of the main gene (ID 0)
        has one copy (and thus should be present) in the configuration `cn`.
    :param kind: Type of the copy-number configuration.
    :param alleles: Allele IDs that have this CN configuration.
    :param description: Human-readable description (e.g. "deletion").
    """

    cn: List[Dict[str, int]]
    kind: CNConfigType
    alleles: Set[str]
    description: str = ""

    @property
    def vector(self):
        return "|".join("".join(str(g[r]) for r in self.cn[0]) for g in self.cn)

    def __str__(self):
        a = " ".join(natsorted(self.alleles))
        return f"CNConfig({str(self.kind)[13:]}; vector={self.vector}; alleles=[{a}])"


@dataclass
class Gene:
    """Gene (and associated pseudogenes) description relative to a reference genome."""

    # Basic information
    name: str
    """Gene name (e.g. _CYP2D6_)."""
    version: str
    """Database version."""
    pharmvar: Optional[str]
    """PharmVar ID."""
    ensembl: Optional[str]
    """ENSEMBL gene ID."""

    # Structure
    refseq: str
    """RefSeq sequence ID."""
    seq: str
    """RefSeq sequence (typically `*1` allele)."""
    genome: str
    """Reference genome version (e.g., hg19)."""
    chr: str
    """Reference genome chromosome."""
    strand: int
    """RefSeq sequence strand within the reference genome."""
    chr_to_ref: Dict[int, int]
    """Position mapping from the reference to the RefSeq sequence."""
    ref_to_chr: Dict[int, int]
    """Position mapping from the RefSeq sequence to the reference."""
    pseudogenes: List[str]
    """Pseudogene names. A pseudogene has ID greater than zero."""
    regions: List[Dict[str, GRange]]
    """
    Collection of regions (names and ranges) for each gene in the database.
    Maps a gene ID to a dictionary that maps a gene region name
    (e.g. `"e9"` for exon 9) to its region in the reference genome.
    Gene 0 is the main gene.
    """
    exons: List[Tuple[int, int]]
    """List of RefSeq exon coordinates."""
    aminoacid: str
    """Aminoacid sequence of the main gene."""

    # Mutations
    mutations: Dict[Tuple[int, str], Tuple[Optional[str], str, int, int, str]]
    """Maps `(position, mutation)` to the corresponding :py:class:`Mutation`."""
    random_mutations: Set[Mutation]
    """Set of mutations that can occur in any allele."""

    do_copy_number: bool
    """Set if the gene is subject to structural variations."""
    cn_configs: Dict[str, CNConfig]
    """
    Copy-number configurations associated with the gene.
    `1` (akin to `*1`) is the default CN configuration (no structural variations
    present).
    """
    unique_regions: List[str]
    """List of genic regions used for copy-number calling."""

    # Alleles
    alleles: Dict[str, MajorAllele]
    """Major star-alleles in the gene."""
    common_tandems: List[Tuple]
    """
    List of common allele tandems. Used in diplotype assignment heuristics.
    For example, the fact that `*13` is always followed by `*1`
    (encoded as `('13', '1')`) will be used to group `*13` and `*1` together
    within the same haplotype (e.g. `*13+*1`).
    """

    def __init__(
        self,
        path: Optional[str],
        name: Optional[str] = None,
        yml: Optional[str] = None,
        genome: Optional[str] = None,
    ) -> None:
        """
        Initialize the object with the database description specified in a YAML file.

        :param path: Location of the YAML file.
        :param name: Gene name.
        :param yml: YML file contents (used if `path` is `None`).
        :param genome: Reference genome version.
        :raises: :py:class:`aldy.common.AldyException`: if the path is not set.
        """

        if path and os.path.exists(path):
            with open(path) as file:
                yml = file.read()
            if not name:
                name = os.path.split(path)[-1].split(".")[0].upper()

        if not yml or not name:
            raise AldyException("Either a path or a name should be given")

        yml = yaml.safe_load(yml)
        self._yml = yml  # for debugging / raw API access
        if genome:
            self.genome = genome
        else:
            self.genome = list(yml["reference"]["mappings"].keys())[0]  # type: ignore
        self._init_basic(yml)
        self._init_regions(yml)
        self._init_alleles(yml)
        self._init_partials()

    def region_at(self, pos: int) -> Optional[Tuple[int, str]]:
        """:returns: Gene ID and a region that covers the position."""
        return self._region_at.get(pos, None)

    def _reverse_op(self, op: str) -> str:
        if ">" in op:
            l, r = op.split(">")
            return f"{rev_comp(l)}>{rev_comp(r)}"
        elif op[:3] == "ins":
            return f"ins{rev_comp(op[3:])}"
        elif op[:3] == "del":
            assert "ins" not in op, "del+ins not yet supported"
            return f"del{rev_comp(op[3:])}"
        return op

    def get_functional(self, mut, infer=True) -> Optional[str]:
        """
        :returns: String describing the mutation effect if a mutation is functional;
            otherwise `None`.
        """
        pos, op = mut
        if (pos, op) in self.mutations:
            return self.mutations[pos, op][0]

        # Calculate based on aminoacid change
        if pos not in self.chr_to_ref:
            return None
        pos = self.chr_to_ref[pos]
        if infer and any(s <= pos < e for s, e in self.exons):
            if ">" not in op:
                return "indel"
            if self.strand < 0:
                op = self._reverse_op(op)
                # pos -= 1
            if op[2] == "N":
                return None
            if self.seq[pos] != op[0]:
                log.warn(f"Bad mutation: {op[0]} != {self.seq[pos]}")
            seq = "".join(
                (
                    self.seq[s:pos] + op[2] + self.seq[pos + 1 : e]
                    if s <= pos < e
                    else self.seq[s:e]
                )
                for s, e in self.exons
            )
            amino = seq_to_amino(seq)
            if amino != self.aminoacid:
                pos = next(i for i, a in enumerate(amino) if a != self.aminoacid[i])
                return f"{self.aminoacid[pos]}{pos + 1}{amino[pos]}"
        return None

    def is_functional(self, mut, infer=True) -> bool:
        """:returns: `True` if a mutation is functional."""
        return self.get_functional(mut, infer) is not None

    def get_rsid(self, *args, default=True) -> str:
        """:returns: rsID if a mutation has it; otherwise `"-"`."""
        assert 1 <= len(args) <= 2
        if len(args) == 2:
            pos, op = args
        else:
            pos, op = args[0]
        res = "-"
        if (pos, op) in self.mutations:
            res = self.mutations[pos, op][1]
        return res if res != "-" or not default else f"{pos + 1}.{op}"

    def get_allele(self, name):
        if name in self.removed:
            name = self.removed[name]
        for an, a in self.alleles.items():
            if name in a.minors:
                return (a, a.minors[name])
        return None

    def get_refseq(self, *args, from_atg=False) -> str:
        """:returns: `True` if a mutation is functional."""
        assert 1 <= len(args) <= 2
        atg_start = self.exons[0][0]
        if len(args) == 2:
            pos, op = args
        else:
            pos, op = args[0]
        if (pos, op) in self.mutations:
            pos, op = self.mutations[pos, op][3:5]
            if from_atg:
                if pos >= atg_start:
                    pos = pos - atg_start + 1
                else:
                    pos = pos - atg_start
            return f"{pos + 1}{op}"
        return "-"

    def deletion_allele(self) -> Optional[str]:
        """
        :returns: The ID of the deletion allele (or `None` if gene has no such allele).
        """

        try:
            return next(
                a
                for a, cn in self.cn_configs.items()
                if cn.kind == CNConfigType.DELETION
            )
        except StopIteration:
            return None

    def has_coverage(self, a: str, pos: int):
        """:returns: `True` if a major allele contains the given mutation."""
        m = self.region_at(pos)
        if m:
            return self.cn_configs[self.alleles[a].cn_config].cn[m[0]][m[1]] > 0
        return False

    def get_wide_region(self):
        """:returns: The genomic region that covers all specified gene regions."""
        mi = min(r.start for g in self.regions for r in g.values())
        ma = max(r.end for g in self.regions for r in g.values())
        return GRange(self.chr, mi, ma)

    def __contains__(self, i: int):
        return i in self.chr_to_ref

    def __getitem__(self, i):
        s, e = self._lookup_range
        if isinstance(i, slice):
            i, j = i.start, i.stop
            if j <= s or i >= e:
                return "N" * (j - i)
            loff = max(0, s - i)
            roff = max(0, j - e)
            if loff:
                i = s
            if roff:
                j = e
            return ("N" * loff) + self._lookup_seq[i - s : j - s] + ("N" * roff)
        else:
            assert isinstance(i, int)
            if not s <= i < e:
                return "N"
            return self._lookup_seq[i - s]

    def __str__(self):
        return f"Gene({self.name})"

    def __repr__(self):
        return f"Gene({self.name})"

    def _init_basic(self, yml) -> None:
        """
        Read basic gene properties (`name`, `seq` and `region`).

        All database YAML coordinates are indexed starting from 1.
        Aldy internally uses 0-based indexing.
        """
        self.name = yml["name"]
        self.version = f"{yml.get('version', 'v?')} ({yml.get('generated', '-')})"
        self.refseq = yml["reference"]["name"]
        self.pharmvar = yml.get("pharmvar", None)
        self.ensembl = yml.get("ensembl", None)
        self.notes = yml.get("notes", "")

        self.seq = yml["reference"]["seq"].replace("\n", "")
        if "patches" in yml["reference"]:
            seq = list(self.seq)
            for [pos, nuc] in yml["reference"]["patches"]:
                seq[pos - 1] = nuc
            self.seq = "".join(seq)

        if self.genome not in yml["reference"]["mappings"]:
            raise AldyException("Gene {self.name} not compatible with {self.genome}")
        self.chr, start, end, strand, cigar = yml["reference"]["mappings"][self.genome]
        self.chr_to_ref = {}
        self.ref_to_chr = {}
        self.strand = 1 if strand == "+" else -1
        pos_ref = 0 if self.strand > 0 else (len(self.seq) - 1)
        pos_chr = start - 1
        for i in cigar.split():
            op, sz = i[0], int(i[1:])
            if op == "M":
                for idx in range(sz):
                    self.chr_to_ref[pos_chr + idx] = pos_ref + idx * self.strand
                    self.ref_to_chr[pos_ref + idx * self.strand] = pos_chr + idx
                pos_chr += sz
                pos_ref += sz * self.strand
            elif op == "I":
                pos_ref += sz * self.strand
            elif op == "D":
                pos_chr += sz
            else:
                raise AldyException("Invalid CIGAR string")

        self._lookup_range = (start - 1, end - 1)
        self._lookup_seq = "".join(
            (
                (
                    rev_comp(self.seq[self.chr_to_ref[i]])
                    if self.strand < 0
                    else self.seq[self.chr_to_ref[i]]
                )
                if i in self.chr_to_ref
                else "N"
            )
            for i in range(*self._lookup_range)
        )

        self.exons = sorted((s - 1, e - 1) for [s, e] in yml["reference"]["exons"])
        self.aminoacid = seq_to_amino("".join(self.seq[s:e] for [s, e] in self.exons))

        self.cpic = {}
        self.cpic_scores = None
        cpic = yml["reference"].get("cpic", {})
        if "scores" in cpic and "conditions" in cpic:
            for fn, s in cpic["conditions"].items():
                self.cpic[fn] = s
            self.cpic_scores = cpic["scores"]
        elif "table" in cpic:
            for fn, ll in cpic["table"].items():
                self.cpic[fn] = {tuple(sorted(i)) for i in ll}

    def _init_regions(self, yml) -> None:
        """
        Calculate the genic regions and pseudogenes (`regions`, `unique_regions`
        and `pseudogenes`). Prepare :py:method:`aldy.gene.Gene.region_at` call.
        """

        self.regions = []
        # Each pseudogene is associated with an index > 0
        self.pseudogenes = []

        for i, g in enumerate(yml["structure"]["genes"]):
            if i > 0:
                self.pseudogenes.append(g)
            regions = {}
            num_exons = 0
            for name, coord in yml["structure"]["regions"][self.genome].items():
                if name[0] == "e" and name[1:].isdigit():  # this is exon
                    num_exons += 1
                if i * 2 + 1 < len(coord) and not coord[i * 2] <= coord[i * 2 + 1]:
                    raise AldyException(
                        f"Malformed YML file {self.name} (structure:regions:{name})"
                    )
                regions[name] = GRange(self.chr, coord[i * 2] - 1, coord[i * 2 + 1] - 1)
            for e in range(1, num_exons):  # fill introns
                r = [regions[f"e{e}"], regions[f"e{e + 1}"]][:: self.strand]
                regions[f"i{e}"] = GRange(self.chr, r[0][2], r[1][1])
            self.regions.append(
                dict(sorted(regions.items(), key=lambda x: x[1])[:: self.strand])
            )

        #: dict[int, (int, str)]:
        #: reverse lookup (gene, region) of gene regions given a location
        #: within the reference genome.
        for d in self.regions:
            regs = sorted((r.start, r.end) for r in d.values())
            for ri in range(1, len(regs)):
                if regs[ri - 1][1] < regs[ri][0]:
                    raise AldyException(
                        "Region {}-{} is not annotated", regs[ri - 1][1], regs[ri][0]
                    )
        self._region_at = {
            i: (g, r)
            for g, d in enumerate(self.regions)
            for r, rng in d.items()
            for i in range(rng.start, rng.end)
        }
        for i in self.chr_to_ref:
            if i not in self._region_at:
                raise AldyException(
                    f"Position {i} not within a named region of {self.name}"
                )
        self.unique_regions = yml["structure"]["cn_regions"]

    def _init_alleles(self, yml) -> None:
        """
        Initialize allele (`alleles`, `common_tandems`) and copy number configurations
        (`cn_configs`). Initialize old notation lookup table (`old_notation`).

        :raise: :py:class:`aldy.common.AldyException` if allele name cannot be found.
        """

        alleles = {}
        fusions_left = {}
        fusions_right = {}
        custom_cn = {}

        # allele ID of the deletion allele (i.e. whole gene is missing).
        deletion_allele = None

        def process_mutation(name, pos, op, info):
            if pos == self.name and op.startswith("deletion:"):
                custom_cn[name] = op[9:].split(",")
            elif pos in self.pseudogenes:
                assert pos == self.pseudogenes[0]  # TODO: relax later
                if op[-1] == "-":
                    fusions_left[name] = op[:-1]
                else:
                    if op[-1] == "+":
                        op = op[:-1]
                    fusions_right[name] = op
            else:
                orig_op, orig_pos = op, pos
                if info == []:
                    info = ["-"]
                rsid, function = info[0], info[1] if len(info) > 1 else None
                if self.strand < 0:
                    if ">" in op:
                        l, r = op.split(">")
                        op = f"{rev_comp(l)}>{rev_comp(r)}"
                        pos = pos + len(l) - 1
                    elif op[:3] == "ins":
                        op = f"ins{rev_comp(op[3:])}"
                        pos += 1
                    elif op[:3] == "del":
                        if "ins" in op[3:]:
                            pd, pi = op[3:].split("ins")
                            op = f"del{rev_comp(pd)}ins{rev_comp(pi)}"
                            pos = pos + len(pd) - 1
                        else:
                            pos = pos + len(op) - 4
                            op = f"del{rev_comp(op[3:])}"
                pos -= 1  # Cast to 0-based index
                if pos not in self.ref_to_chr:
                    log.warn(f"Ignoring {pos}.{op} in {name} (not in {self.refseq})")
                elif self.region_at(self.ref_to_chr[pos]) is None:
                    log.warn(f"Ignoring {pos}.{op} in {name} (not in named region)")
                else:
                    self.mutations.setdefault(
                        (self.ref_to_chr[pos], op),
                        (function, rsid, pos, orig_pos - 1, orig_op),
                    )
                    yield Mutation(self.ref_to_chr[pos], op)

        self.mutations = {}
        self.yml_mutations = yml["alleles"]
        self.random_mutations = set()
        for pos, op, *info in yml["alleles"].get("random", []):
            if isinstance(pos, str) and pos == "ignored":
                continue
            for m in process_mutation("random", pos, op, info):
                self.random_mutations.add(m)
        self.mutation_groups = {}
        for name, muts in yml["alleles"].get("groups", {}).items():
            self.mutation_groups[name] = set()
            for pos, op, *info in muts:
                for m in process_mutation(name, pos, op, info):
                    self.mutation_groups[name].add(m)

        for name, allele in yml["alleles"].items():
            if name in ["random", "groups"]:
                continue
            if allele.get("ignored", False):
                continue
            name = allele_name(name)
            mutations: Set[Mutation] = set()
            if [self.name, "deletion"] in allele["mutations"]:
                deletion_allele = name
            else:
                for pos, op, *info in allele["mutations"]:
                    if isinstance(pos, str) and pos == "ignored":
                        continue
                    if pos == self.name and op in self.mutation_groups:
                        log.warn(f"Ignoring {op} in {name}")
                        continue
                    for m in process_mutation(name, pos, op, info):
                        mutations.add(m)

            alt_name = allele.get("label", None)
            if alt_name:
                alt_name = allele_name(alt_name)
            alleles[name] = MinorAllele(
                name,
                alt_name,
                mutations,
                allele.get("activity", "unknown"),
                allele.get("evidence", None),
                allele.get("pharmvar", None),
            )

        self.common_tandems: List[tuple] = []
        if "tandems" in yml["structure"]:
            self.common_tandems = [tuple(y) for y in yml["structure"]["tandems"]]

        # Set copy number configurations
        # TODO: currently fusions only cover the space between the main gene and
        # the first pseudogene. Multi-pseudogene fusions are not supported.

        self.do_copy_number = bool(
            deletion_allele or len(fusions_left) or len(fusions_right) or len(custom_cn)
        )
        self.cn_configs: Dict[str, CNConfig] = dict()

        for i, _ in enumerate(self.pseudogenes):
            if list(self.regions[i + 1].keys()) != list(self.regions[0].keys()):
                raise AldyException("Invalid database structure")
        rank = {a: ai for ai, a in enumerate(self.regions[0])}
        inverse_cn: Dict[tuple, str] = dict()
        # Left fusions are PSEUDOGENE + GENE fusions
        for a, brk in fusions_left.items():
            cn = [
                {r: int(rank[r] >= rank[brk]) for r in self.regions[0]},
                {r: int(rank[r] < rank[brk]) for r in self.regions[1]},
            ]
            key = freezekey(cn)
            if key not in inverse_cn:
                self.cn_configs[a] = CNConfig(
                    cn,
                    CNConfigType.LEFT_FUSION,
                    {a},
                    f"{self.pseudogenes[0]} fusion until {brk}",
                )
                inverse_cn[key] = a
            else:
                self.cn_configs[inverse_cn[key]].alleles.add(a)
        # Deletion is a special kind of left fusion
        if deletion_allele is not None:
            cn = [{r: 0 for r in self.regions[0]}]
            if len(self.pseudogenes) > 0:
                cn.append({r: 1 for r in self.regions[1]})
            self.cn_configs[deletion_allele] = CNConfig(
                cn,
                CNConfigType.DELETION,
                {deletion_allele},
                f"{self.name} deletion",
            )
        # Right fusions GENE + PSEUDOGENE + whole copy of PSEUDOGENE fusions
        for a, brk in fusions_right.items():
            cn = [
                {r: int(rank[r] < rank[brk]) for r in self.regions[0]},
                {r: 1 + int(rank[r] >= rank[brk]) for r in self.regions[1]},
            ]
            key = freezekey(cn)
            if key not in inverse_cn:
                self.cn_configs[a] = CNConfig(
                    cn,
                    CNConfigType.RIGHT_FUSION,
                    {a},
                    f"{self.pseudogenes[0]} conservation after {brk}",
                )
                inverse_cn[key] = a
            else:
                self.cn_configs[inverse_cn[key]].alleles.add(a)
        # Custom deletions (main gene only)
        for a, items in custom_cn.items():
            cn = [{r: int(r not in items) for r in self.regions[0]}]
            if len(self.regions) > 1:
                cn.append({r: 1 for r in self.regions[1]})
            key = freezekey(cn)
            if key not in inverse_cn:
                self.cn_configs[a] = CNConfig(
                    cn,
                    CNConfigType.CUSTOM,
                    {a},
                    f"{self.name} deletion in {','.join(items)}",
                )
                inverse_cn[key] = a
            else:
                self.cn_configs[inverse_cn[key]].alleles.add(a)

        # Normal CN case
        used_alleles = {a for cn in self.cn_configs.values() for a in cn.alleles}
        has_pseudogenes = len(self.pseudogenes) > 0
        default_cn = [
            {r: 1 for r in self.regions[g]} for g in range(1 + has_pseudogenes)
        ]
        self.cn_configs = {min(v.alleles): v for v in self.cn_configs.values()}
        self.cn_configs["1"] = CNConfig(
            default_cn,
            CNConfigType.DEFAULT,
            set(alleles.keys()) - used_alleles,
            "Standard copy-number configuration",
        )

        # Set CN of "empty" regions to zero (e.g. CYP2D6.pce)
        for _, conf in self.cn_configs.items():
            for g, r in enumerate(conf.cn):
                for rg in r:
                    if self.regions[g][rg].end - self.regions[g][rg].start <= 0:
                        r[rg] = 0

        # Set up major and minor allele structures
        alleles_inverse: Dict[tuple, set] = collections.defaultdict(set)
        for a in alleles:
            cn_config = next(
                cn for cn, conf in self.cn_configs.items() if a in conf.alleles
            )
            fn_muts = sorted(
                m for m in alleles[a].neutral_muts if self.is_functional(m)
            )
            alleles_inverse[cn_config, tuple(fn_muts)].add(a)

        # Ensure that each distinct major allele has unique prefix
        used_names: Dict[str, int] = {}
        new_names = {}
        changed_cn_configs = {}
        for key, minors in alleles_inverse.items():
            an = min(minors)
            name = an.split(".")[0]  # remove ".0XX" suffix
            if name in used_names:  # special case
                name = alleles[an].alt_name if alleles[an].alt_name else an
            if name in used_names:  # Append letter
                used_names[name] += 1
                log.debug(f"[gene] renaming {name} -> {name}:{used_names[name]}")
                name += f":{used_names[name]}"
                used_names[name] = 1
            else:
                used_names[name] = 1
            if an in self.cn_configs and an != name:
                self.cn_configs[name] = self.cn_configs[an]
                changed_cn_configs[an] = name
                del self.cn_configs[an]
            new_names[key] = name
        self.alleles: Dict[str, MajorAllele] = dict()
        for key, minors in alleles_inverse.items():
            self.alleles[new_names[key]] = MajorAllele(
                name=new_names[key],
                cn_config=changed_cn_configs.get(key[0], key[0]),
                func_muts=set(key[1]),
                minors={
                    sa: MinorAllele(
                        sa,
                        alleles[sa].alt_name,
                        set(alleles[sa].neutral_muts) - set(key[1]),
                        alleles[sa].activity,
                        alleles[sa].evidence,
                        alleles[sa].pharmvar,
                    )
                    for sa in natsorted(minors)
                },
            )

    def _init_partials(self) -> None:
        """
        Construct partial major alleles.
        If a major allele is cut in half by a fusion, we will create a new major
        allele that contains the functional mutations that have survived the fusion
        event.

        .. note::
            This is currently supported only for left fusions.
        """

        def preserved_mutations(f, m):
            def filter_f(m):
                m = self.region_at(m.pos)
                if m:
                    return self.cn_configs[f].cn[m[0]][m[1]] > 0
                return False

            return set(filter(filter_f, m))

        for f in filter(
            lambda x: self.cn_configs[x].kind == CNConfigType.LEFT_FUSION,
            self.cn_configs,
        ):
            # Do not extend left fusions that are already defined
            # by some functional mutations
            # HACK: This is hacky but works well for now
            if len(self.alleles[f].func_muts) > 0:
                continue
            add: Dict[tuple, MajorAllele] = {}
            for an, a in self.alleles.items():
                # We only make partial major alleles from non-fused major alleles
                if a.cn_config != "1":
                    continue
                new_name = f"{f}#{an}"
                new_muts = preserved_mutations(f, a.func_muts)
                key = sorted_tuple(new_muts)
                if key in add:
                    add[key].minors.update(
                        {
                            f"{f}#{san}": MinorAllele(
                                f"{f}#{san}",
                                None,
                                preserved_mutations(f, sa.neutral_muts),
                                sa.activity,
                                sa.evidence,
                                sa.pharmvar,
                            )
                            for san, sa in a.minors.items()
                        }
                    )
                else:
                    add[key] = MajorAllele(
                        name=new_name,
                        cn_config=f,
                        func_muts=new_muts,
                        minors={
                            f"{f}#{san}": MinorAllele(
                                f"{f}#{san}",
                                neutral_muts=preserved_mutations(f, sa.neutral_muts),
                            )
                            for san, sa in a.minors.items()
                        },
                    )

            # Remove fusion (will be replaced at least by allele "1#{f}")
            del self.alleles[f]
            self.alleles.update({a.name: a for a in add.values()})

        self.removed = {}
        for an, a in self.alleles.items():
            # Clean up minor alleles (as many might be identical after a fusion).
            # Put a reference to the cleaned-up alleles in `alt_name` field.
            minors: Dict[tuple, List[str]] = collections.defaultdict(list)
            for s in a.minors:
                key = sorted_tuple(a.minors[s].neutral_muts)
                minors[key].append(s)
            for sa in minors.values():
                if len(sa) > 1:
                    for s in sa:
                        if s != min(sa) and "#" not in s:
                            self.removed[s] = min(sa)
                            log.debug(
                                f"Removing {self.name}*{s} as it is the same as"
                                f" {min(sa)}"
                            )
            self.alleles[an] = MajorAllele(
                self.alleles[an].name,
                self.alleles[an].cn_config,
                self.alleles[an].func_muts,
                {
                    min(sa): MinorAllele(
                        min(sa),
                        a.minors[min(sa)].alt_name,
                        set(nm),
                        a.minors[min(sa)].activity,
                        a.minors[min(sa)].evidence,
                        a.minors[min(sa)].pharmvar,
                    )
                    for nm, sa in minors.items()
                },
            )

        # TODO: prune identical post-fusion alleles
        # (e.g. *2 after *13A and *13B is the same--- no need to have 2 items).

        for _, v in self.cn_configs.items():
            v.alleles.clear()
        for a in self.alleles.values():
            self.cn_configs[a.cn_config].alleles.add(a.name)



# 786
# X source: test_cn_real.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.

# flake8: noqa

import pytest  # noqa
import re
import subprocess
import platform
import datetime

from tempfile import NamedTemporaryFile as tmpfile

from aldy.__main__ import get_version, main
from aldy.common import script_path, log
from aldy.version import __version__


YEAR = datetime.datetime.now().year
HEADER = f"""
🐿  Aldy v{__version__} (Python {platform.python_version()} on {get_version()})
   (c) 2016-{datetime.datetime.now().year} Aldy Authors. All rights reserved.
   Free for non-commercial/academic use only.
""".strip()


def escape_ansi(line):
    """
    Inspired by
    https://www.tutorialspoint.com/How-can-I-remove-the-ANSI-escape-sequences-from-a-string-in-python
    """  # noqa
    return re.compile(r"(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]").sub("", line)


def assert_file(monkeypatch, file, solver, expected, params=None, warn=False):
    lines = []

    def log_info(*args):
        nonlocal lines

        s = str.format(*args)
        lines += s.split("\n")  # type: ignore

    monkeypatch.setattr(log, "info", log_info)
    if warn:
        monkeypatch.setattr(log, "warn", log_info)

    args = {
        **{
            "--gene": "cyp2d6",
            "--profile": "illumina",
            "--param": "threshold=0.5",
            "--param": "gap=0",
            "--param": "max-minor-solutions=3",
            "--param": "minor_phase_vars=10",
            "--solver": solver,
        },
        **(params or {}),
    }
    main(["genotype", file] + [i for k, v in args.items() for i in [k, v]])
    expected = "\n".join(e.strip() for e in expected.strip().split("\n"))
    lines = "\n".join(escape_ansi(l).strip() for l in lines).strip()
    assert lines == expected


EXPECTED_NA10860 = f"""
{HEADER}
Genotyping sample NA10860.bam...
Potential CYP2D6 gene structures for NA10860:
   1: 2x*1,1x*141.001X (confidence: 100%)
   2: 2x*1,1x*61 (confidence: 100%)
Potential major CYP2D6 star-alleles for NA10860:
   1: 1x*1, 1x*4, 1x*4.021X (confidence: 100%)
   2: 1x*1, 1x*4.021, 1x*4NX (confidence: 100%)
   3: 1x*4, 1x*4NX, 1x*139 (confidence: 100%)
   4: 1x*4.021, 1x*4J, 1x*61 (confidence: 100%)
   5: 1x*4.021, 1x*4J, 1x*83X (confidence: 100%)
   6: 1x*4.021, 1x*4M, 1x*36X (confidence: 100%)
   7: 1x*4.021X, 1x*4C, 1x*185 (confidence: 100%)
   8: 1x*4.021X, 1x*4J, 1x*39 (confidence: 100%)
   9: 1x*4.021X, 1x*4M, 1x*10 (confidence: 100%)
   10: 1x*4.021X2, 1x*4NX, 1x*74 (confidence: 100%)
Potential CYP2D6 star-alleles for NA10860:
   1: *1 / *4.021 + *4NX (confidence=100%)
      Minor alleles: *(1.018 +rs111564371 +rs112568578 +rs113889384 +rs28371713), *(4.013X -rs28371738), *4.021
   2: *1 / *4 + *4.021X (confidence=100%)
      Minor alleles: *(1.018 +rs111564371 +rs112568578 +rs113889384 +rs28371713), *4.001, *(4.021X -rs28371738)
CYP2D6 results:
  - *1 / *4.021 + *4NX
    Minor: [*1.018 +rs111564371 +rs112568578 +rs113889384 +rs28371713] / [*4.021] + [*4.013X -rs28371738]
    Legacy notation: [*1.018 +rs111564371 +rs112568578 +rs113889384 +rs28371713] / [*4.021] + [*4NX -rs28371738]
    CPIC functionality: intermediate (activity score = 1.0)
    Estimated activity for *1: normal (score: 1.0); evidence: D (see https://www.pharmvar.org/haplotype/662 for details)
    Estimated activity for *4.021: none (score: 0.0); evidence: D (see https://www.pharmvar.org/haplotype/2566 for details)
    Estimated activity for *4NX: none (score: 0.0); evidence: D (see https://www.pharmvar.org/haplotype/2559 for details)
  - *1 / *4 + *4.021X
    Minor: [*1.018 +rs111564371 +rs112568578 +rs113889384 +rs28371713] / [*4.001] + [*4.021X -rs28371738]
    Legacy notation: [*1.018 +rs111564371 +rs112568578 +rs113889384 +rs28371713] / [*4A] + [*4.021X -rs28371738]
    CPIC functionality: intermediate (activity score = 1.0)
    Estimated activity for *1: normal (score: 1.0); evidence: D (see https://www.pharmvar.org/haplotype/662 for details)
    Estimated activity for *4: none (score: 0.0); evidence: D (see https://www.pharmvar.org/haplotype/2557 for details)
    Estimated activity for *4.021X: none (score: 0.0); evidence: D (see https://www.pharmvar.org/haplotype/2566 for details)
"""


def test_NA10860(monkeypatch, solver):
    file = script_path("aldy.tests.resources/NA10860.bam")
    assert_file(monkeypatch, file, solver, EXPECTED_NA10860)


def test_NA10860_debug(monkeypatch, solver):
    expected_tar = """
    ./
    ./NA10860.CYP2D6.cn.lp
    ./NA10860.CYP2D6.dump
    ./NA10860.CYP2D6.genome
    ./NA10860.CYP2D6.major0.lp
    ./NA10860.CYP2D6.major1.lp
    ./NA10860.log
    ./NA10860.yml
    """
    file = script_path("aldy.tests.resources/NA10860.bam")
    with tmpfile(suffix=".tar.gz") as tmp:
        with tmpfile(mode="w") as out, tmpfile(mode="w") as out_log:
            out.close()
            out_log.close()
            assert_file(
                monkeypatch,
                file,
                solver,
                EXPECTED_NA10860 + "Preparing debug archive...",
                {"--debug": tmp.name[:-7], "--log": out_log.name, "--output": out.name},
            )
            with open(script_path("aldy.tests.resources/NA10860.out.expected")) as f:
                expected = f.read()
            with open(out.name) as f:
                produced = f.read()
            assert produced == expected
            # Check logs
            with open(out_log.name) as f:
                log = f.read()
            s = "    rs1058172    42523528.C>T    3267G>A    "
            s += "(cov=  17, cn= 0.8; region=e7; impact=R365H; )\n"
            s = "[major] status= optimal; opt= 1.47; solution= 1x*1, 1x*4, 1x*4.021"
            assert s in log
        out = subprocess.check_output(f"tar tzf {tmp.name}", shell=True).decode("utf-8")
        out = "\n".join(i.strip() for i in sorted(out.strip().split("\n")))
        assert out == "\n".join(i.strip() for i in expected_tar.strip().split("\n"))


def test_NA10860_hg38(monkeypatch, solver):
    file = script_path("aldy.tests.resources/NA10860_hg38.bam")
    assert_file(
        monkeypatch, file, solver, EXPECTED_NA10860.replace("NA10860", "NA10860_hg38")
    )


def test_NA10860_gap(monkeypatch, solver):
    expected = f"""
    {HEADER}
    Genotyping sample NA10860.bam...
    Potential CYP2D6 gene structures for NA10860:
    1: 2x*1,1x*141.001X (confidence: 100%)
    2: 2x*1,1x*61 (confidence: 100%)
    3: 1x*1,1x*61,1x*141.001X (confidence: 90%)
    4: 1x*1,2x*141.001X (confidence: 90%)
    5: 1x*1,2x*61 (confidence: 90%)
    6: 3x*1 (confidence: 90%)
    Potential major CYP2D6 star-alleles for NA10860:
    1: 1x*1, 1x*4, 1x*4.021X (confidence: 100%)
    2: 1x*1, 1x*4.021, 1x*4NX (confidence: 100%)
    3: 1x*4, 1x*4NX, 1x*139 (confidence: 100%)
    4: 1x*4.021, 1x*4J, 1x*61 (confidence: 100%)
    5: 1x*4.021, 1x*4J, 1x*83X (confidence: 100%)
    6: 1x*4.021, 1x*4M, 1x*36X (confidence: 100%)
    7: 1x*4.021X, 1x*4C, 1x*185 (confidence: 100%)
    8: 1x*4.021X, 1x*4J, 1x*39 (confidence: 100%)
    9: 1x*4.021X, 1x*4M, 1x*10 (confidence: 100%)
    10: 1x*4.021X2, 1x*4NX, 1x*74 (confidence: 100%)
    Potential CYP2D6 star-alleles for NA10860:
    1: *1 / *4 + *4.021X (confidence=100%)
        Minor alleles: *(1.018 +rs28371713), *(4.001 +rs111564371 +rs112568578 +rs113889384), *(4.021X -rs28371738)
    2: *1 / *4.021 + *4NX (confidence=100%)
        Minor alleles: *(1.018 +rs28371713), *(4.013X -rs28371738), *(4.021 +rs111564371 +rs112568578 +rs113889384)
    CYP2D6 results:
    - *1 / *4 + *4.021X
        Minor: [*1.018 +rs28371713] / [*4.001 +rs111564371 +rs112568578 +rs113889384] + [*4.021X -rs28371738]
        Legacy notation: [*1.018 +rs28371713] / [*4A +rs111564371 +rs112568578 +rs113889384] + [*4.021X -rs28371738]
        CPIC functionality: intermediate (activity score = 1.0)
        Estimated activity for *1: normal (score: 1.0); evidence: D (see https://www.pharmvar.org/haplotype/662 for details)
        Estimated activity for *4: none (score: 0.0); evidence: D (see https://www.pharmvar.org/haplotype/2557 for details)
        Estimated activity for *4.021X: none (score: 0.0); evidence: D (see https://www.pharmvar.org/haplotype/2566 for details)
    - *1 / *4.021 + *4NX
        Minor: [*1.018 +rs28371713] / [*4.021 +rs111564371 +rs112568578 +rs113889384] + [*4.013X -rs28371738]
        Legacy notation: [*1.018 +rs28371713] / [*4.021 +rs111564371 +rs112568578 +rs113889384] + [*4NX -rs28371738]
        CPIC functionality: intermediate (activity score = 1.0)
        Estimated activity for *1: normal (score: 1.0); evidence: D (see https://www.pharmvar.org/haplotype/662 for details)
        Estimated activity for *4.021: none (score: 0.0); evidence: D (see https://www.pharmvar.org/haplotype/2566 for details)
        Estimated activity for *4NX: none (score: 0.0); evidence: D (see https://www.pharmvar.org/haplotype/2559 for details)
    """
    file = script_path("aldy.tests.resources/NA10860.bam")
    assert_file(monkeypatch, file, solver, expected, {"--param": "gap=0.3"})


def test_NA10860_cn(monkeypatch, solver):
    expected = f"""
    {HEADER}
    Genotyping sample NA10860.bam...
    Potential CYP2D6 gene structures for NA10860:
    1: 2x*1 (confidence: 100%)
    Potential major CYP2D6 star-alleles for NA10860:
    1: 1x*1, 1x*4.021 (confidence: 100%)
    2: 1x*4, 1x*139 (confidence: 100%)
    3: 1x*4.021X2, 1x*74 (confidence: 100%)
    Best CYP2D6 star-alleles for NA10860:
    1: *1 / *4.021 (confidence=100%)
        Minor alleles: *(1.018 +rs113889384 +rs28371713), *4.021
    CYP2D6 results:
    - *1 / *4.021
        Minor: [*1.018 +rs113889384 +rs28371713] / [*4.021]
        Legacy notation: [*1.018 +rs113889384 +rs28371713] / [*4.021]
        CPIC functionality: intermediate (activity score = 1.0)
        Estimated activity for *1: normal (score: 1.0); evidence: D (see https://www.pharmvar.org/haplotype/662 for details)
        Estimated activity for *4.021: none (score: 0.0); evidence: D (see https://www.pharmvar.org/haplotype/2566 for details)
    """
    file = script_path("aldy.tests.resources/NA10860.bam")
    assert_file(monkeypatch, file, solver, expected, {"--cn": "1,1"})


def test_NA10860_vcf_out(monkeypatch, solver):
    file = script_path("aldy.tests.resources/NA10860.bam")
    with tmpfile(suffix=".vcf", mode="w") as out:
        out.close()
        assert_file(monkeypatch, file, solver, EXPECTED_NA10860, {"--output": out.name})
        with open(script_path("aldy.tests.resources/NA10860.vcf.expected")) as f:
            expected = f.read()
        with open(out.name) as f:
            produced = f.read()
        assert produced == expected.replace("aldy-v4.8", f"aldy-v{__version__}")


def test_fusion(monkeypatch, solver):
    # NA19785 pgrnseq-v1
    expected = f"""
    {HEADER}
    Genotyping sample HARD.dump.tar.gz...
    Potential CYP2D6 gene structures for NA19785:
    1: 2x*1,1x*79 (confidence: 100%)
    Potential major CYP2D6 star-alleles for NA19785:
    1: 1x*1, 1x*2, 1x*79#2 (confidence: 100%)
    2: 1x*2, 1x*34, 1x*79#10 (confidence: 100%)
    3: 2x*2, 1x*79#1 (confidence: 100%)
    Potential CYP2D6 star-alleles for NA19785:
    1: *1 / *79 + *2 (confidence=100%)
        Minor alleles: *1, *2, *79#11
    2: *2 / *79 + *2 (confidence=100%)
        Minor alleles: *2, *2, *79#1
    3: *34 / *79 + *2 (confidence=100%)
        Minor alleles: *2, *34, *79#10
    CYP2D6 results:
    - *1 / *79 + *2
        Minor: [*1] / [*79#11] + [*2]
        Legacy notation: [*1] / [*79#11] + [*2]
        Estimated activity for *1: unknown
        Estimated activity for *2: unknown
        Estimated activity for *79#2: unknown
    - *2 / *79 + *2
        Minor: [*2] / [*79#1] + [*2]
        Legacy notation: [*2] / [*79#1] + [*2]
        Estimated activity for *2: unknown
        Estimated activity for *79#1: unknown
        Estimated activity for *2: unknown
    - *34 / *79 + *2
        Minor: [*34] / [*79#10] + [*2]
        Legacy notation: [*34] / [*79#10] + [*2]
        Estimated activity for *2: unknown
        Estimated activity for *34: unknown
        Estimated activity for *79#10: unknown
    """
    file = script_path("aldy.tests.resources/HARD.dump.tar.gz")
    gene = script_path("aldy.tests.resources/cyp2d6.test.yml")
    assert_file(monkeypatch, file, solver, expected, {"--gene": gene})


def test_fusion_off(monkeypatch, solver):
    expected = f"""
    {HEADER}
    Genotyping sample HARD.dump.tar.gz...
    Potential CYP2D6 gene structures for NA19785:
    1: 3x*1 (confidence: 100%)
    Potential major CYP2D6 star-alleles for NA19785:
    1: 1x*1, 2x*2 (confidence: 100%)
    Best CYP2D6 star-alleles for NA19785:
    1: *1 / *2 + *2 (confidence=100%)
        Minor alleles: *1, *2, *2
    CYP2D6 results:
    - *1 / *2 + *2
        Minor: [*1] / [*2] + [*2]
        Legacy notation: [*1] / [*2] + [*2]
        Estimated activity for *1: unknown
        Estimated activity for *2: unknown
        Estimated activity for *2: unknown
    """
    file = script_path("aldy.tests.resources/HARD.dump.tar.gz")
    gene = script_path("aldy.tests.resources/cyp2d6.test.yml")
    assert_file(
        monkeypatch,
        file,
        solver,
        expected,
        {"--param": "cn-fusion-left=10", "--gene": gene},
    )


def test_NA07000_vcf_in(monkeypatch, solver):
    expected = f"""
    {HEADER}
    Genotyping sample NA07000_SLCO1B1.vcf.gz...
    WARNING: Cannot detect genome, defaulting to hg19.
    WARNING: Using VCF file. Copy-number calling is not available.
    Using VCF sample NA07000
    Potential SLCO1B1 gene structures for NA07000:
    1: 2x*1 (confidence: 100%)
    Potential major SLCO1B1 star-alleles for NA07000:
    1: 1x*1, 1x*15 (confidence: 100%)
    2: 1x*5, 1x*37 (confidence: 100%)
    Best SLCO1B1 star-alleles for NA07000:
    1: *1 / *15 (confidence=100%)
        Minor alleles: *1.003, *15.001
    SLCO1B1 results:
    - *1 / *15
        Minor: [*1.003] / [*15.001]
        Legacy notation: [*1.003] / [*15A, SLCO1B1*15B, SLCO1B1*17]
        CPIC functionality: intermediate
        Estimated activity for *1: normal; evidence: D (see https://www.pharmvar.org/haplotype/1713 for details)
        Estimated activity for *15: none; evidence: D (see https://www.pharmvar.org/haplotype/1704 for details)
    """
    file = script_path("aldy.tests.resources/NA07000_SLCO1B1.vcf.gz")
    assert_file(
        monkeypatch,
        file,
        solver,
        expected,
        {"--param": "max-minor-solutions=1", "--gene": "SLCO1B1"},
        warn=True,
    )


def test_fix_insertions(monkeypatch, solver):
    expected = f"""
    {HEADER}
    Genotyping sample INS.dump.tar.gz...
    Potential CYP2D6 gene structures for NA17102:
    1: 2x*1 (confidence: 100%)
    Potential major CYP2D6 star-alleles for NA17102:
    1: 1x*1, 1x*40 (confidence: 100%)
    Best CYP2D6 star-alleles for NA17102:
    1: *1 / *40 (confidence=100%)
        Minor alleles: *1, *40
    CYP2D6 results:
    - *1 / *40
        Minor: [*1] / [*40]
        Legacy notation: [*1] / [*40]
        Estimated activity for *1: unknown
        Estimated activity for *40: unknown
    """
    file = script_path("aldy.tests.resources/INS.dump.tar.gz")
    gene = script_path("aldy.tests.resources/cyp2d6.test.yml")
    assert_file(
        monkeypatch,
        file,
        solver,
        expected,
        {"--gene": gene, "--param": "max-minor-solutions=1"},
    )


def test_profile(monkeypatch, capsys):
    lines = []

    def log_info(*args):
        nonlocal lines

        s = str.format(*args)
        lines += s.split("\n")  # type: ignore

    monkeypatch.setattr(log, "info", log_info)

    expected = f"""
    {HEADER}
    Scanning 1:60355979-201113425...
    Scanning 2:234492389-234684945...
    Scanning 4:69916092-89158474...
    Scanning 6:18125541-18161374...
    Scanning 7:1016834-117359025...
    Scanning 8:18021970-18261723...
    Scanning 10:96516437-135355620...
    Scanning 11:14896554-67357124...
    Scanning 12:21278127-21395730...
    Scanning 13:48605702-48629878...
    Scanning 15:75008882-75051948...
    Scanning 16:31099174-31112276...
    Scanning 19:15985833-41716444...
    Scanning 22:19923262-42549249...
    Scanning MT:1-2700...
    Scanning X:153756605-153781787...
    """
    main(["profile", script_path("aldy.tests.resources/NA10860.bam")])
    lines = "\n".join(escape_ansi(l).strip() for l in lines).strip()
    expected = "\n".join(e.strip() for e in expected.strip().split("\n"))
    assert lines == expected

    captured = capsys.readouterr()
    with open(script_path("aldy.tests.resources/NA10860.profile")) as f:
        expected = f.read()
    assert captured.out == expected

    lines = []
    expected = f"""
    {HEADER}
    Scanning 1:59890307-201118566...
    Scanning 2:233583743-233776299...
    Scanning 4:69050374-88237322...
    Scanning 6:18125310-18161143...
    Scanning 7:977198-117718971...
    Scanning 8:18164461-18404213...
    Scanning 10:94756680-133542116...
    Scanning 11:14875008-67589653...
    Scanning 12:21125193-21242796...
    Scanning 13:48031566-48055742...
    Scanning 15:74716541-74759607...
    Scanning 16:31087853-31100955...
    Scanning 19:15875023-41210539...
    Scanning 22:19935739-42153258...
    Scanning MT:1-2700...
    Scanning X:154528390-154553572...
    """
    main(
        ["profile", script_path("aldy.tests.resources/NA10860.bam"), "--genome", "hg38"]
    )
    lines = "\n".join(escape_ansi(l).strip() for l in lines).strip()
    expected = "\n".join(e.strip() for e in expected.strip().split("\n"))
    assert lines == expected

    captured = capsys.readouterr()
    with open(script_path("aldy.tests.resources/NA10860.profile.hg38")) as f:
        expected = f.read()
    assert captured.out == expected


def test_pacbio(monkeypatch, solver):
    expected = f"""
    {HEADER}
    Genotyping sample HG03166.pb.bam...
    Potential CYP2D6 gene structures for HG03166:
    1: 2x*1 (confidence: 100%)
    Potential major CYP2D6 star-alleles for HG03166:
    1: 1x*2, 1x*40 (confidence: 100%)
    Best CYP2D6 star-alleles for HG03166:
    1: *2 / *40 (confidence=100%)
        Minor alleles: *2.023, *40.001
    CYP2D6 results:
    - *2 / *40
        Minor: [*2.023] / [*40.001]
        Legacy notation: [*2.023] / [*40]
        CPIC functionality: intermediate (activity score = 1.0)
        Estimated activity for *2: normal (score: 1.0); evidence: D (see https://www.pharmvar.org/haplotype/2549 for details)
        Estimated activity for *40: none (score: 0.0); evidence: D (see https://www.pharmvar.org/haplotype/231 for details)
    """
    file = script_path("aldy.tests.resources/HG03166.pb.bam")
    assert_file(
        monkeypatch, file, solver, expected, {"--profile": "pacbio-hifi-targeted"}
    )


def test_load_all_genes(monkeypatch, solver):
    import importlib_resources
    from aldy.gene import Gene
    from aldy.profile import Profile
    from aldy.common import AldyException

    avail_genes = importlib_resources.files("aldy.resources.genes").iterdir()
    for g in sorted(avail_genes):
        if g.is_file() and g.suffix == ".yml":
            print(g)
            for genome in ["hg19", "hg38"]:
                gx = Gene(g, genome=genome)
                p = Profile.load(gx, "illumina")
                for gene, gr in enumerate(gx.regions):
                    for region in gr:
                        assert region in p.data[gx.name]
                        assert sum(p.data[gx.name][region]) > 0



# 786
# Aldy source: test_cn_synthetic.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


import pytest  # noqa

from aldy.profile import Profile
from aldy.cn import solve_cn_model
from aldy.common import SOLUTION_PRECISION


def assert_cn(gene, solver, expected, cov, expected_obj=None, gap=0.0):
    profile = Profile("test")
    profile.gap = gap
    sols = solve_cn_model(
        gene,
        profile,
        cn_configs=gene.cn_configs,
        max_cn=profile.cn_max,
        region_coverage=cov,
        solver=solver,
    )
    if expected_obj:
        for s in sols:
            print(s.score, expected_obj)
            assert abs(s.score - expected_obj) < SOLUTION_PRECISION, "Score mismatch"
    best_sol_score = min(sols, key=lambda s: s.score).score
    for s in sols:
        assert (
            s.score - best_sol_score * (1 + gap)
        ) < SOLUTION_PRECISION, "Gap violation"
    sols = sorted([sorted(s.solution.items()) for s in sols])
    expected = sorted([sorted(s.items()) for s in expected])
    print(sols)
    print(expected)
    assert sols == expected, "Solution mismatch"


def make_coverage(gene, lst):
    cov = {}
    for r in gene.unique_regions:
        cov[r] = next(lst)
    return cov


def test_basic(toy_gene, solver):
    # Test two copies of *1
    assert_cn(
        toy_gene,
        solver,
        [{"1": 2}],
        make_coverage(toy_gene, zip([2, 2, 2, 2, 2], [2, 2, 2, 2, 2])),
        0 + 0 + (7.5 / len(toy_gene.unique_regions)),
    )
    # Test two copies of *1 with slightly perturbed coverage
    g1 = [1.8, 2.2, 2.1, 1.9, 1.7]
    g2 = [2.05, 1.95, 2, 2, 2.7]
    d1 = 2 * sum(abs(a - b) / (max(a, b) + 1) for a, b in zip(g1, g2))
    d2 = sum(abs(a - 2) for a in g1) / len(toy_gene.unique_regions)
    assert_cn(
        toy_gene,
        solver,
        [{"1": 2}],
        make_coverage(toy_gene, zip(g1, g2)),
        d1 + d2 + (7.5 / len(toy_gene.unique_regions)),
    )


def test_deletion(toy_gene, solver):
    # Test a single copy of *1 (*6 is deletion allele)
    assert_cn(
        toy_gene,
        solver,
        [{"1": 1}],
        make_coverage(toy_gene, zip([1, 1, 1, 1, 1], [2, 2, 2, 2, 2])),
        0 + 0 + (7.5 / len(toy_gene.unique_regions)),
    )
    # Test whole gene deletion
    assert_cn(
        toy_gene,
        solver,
        [{}],
        make_coverage(toy_gene, zip([0, 0, 0, 0, 0], [2, 2, 2, 2, 2])),
        0 + 0 + (7.5 / len(toy_gene.unique_regions)),
    )
    # TODO: Test 2 deletions with no coverage
    # (needs special handling as deletions imply pseudogene coverage)
    # assert_cn(toy_gene,solver,
    #           [{}],
    #           make_coverage(toy_gene, zip([0,0, 0,0, 0], [0,0, 0,0, 0])),
    #           2 + 0 + (7.5 / len(toy_gene.unique_regions)))


def test_left_fusion(toy_gene, solver):
    base = 7.5 / len(toy_gene.unique_regions)

    # Test two fused copies (*4 is defined as 00011|11100)
    assert_cn(
        toy_gene,
        solver,
        [{"4": 2}],
        make_coverage(toy_gene, zip([0, 0, 0, 2, 2], [2, 2, 2, 0, 0])),
        0 + 0 + 0.5 * ((base * 1.5) + (base * 1.5)),
    )
    # Test one fused and one normal (*1) allele
    # Note: each left fusion operates on the whole genic region;
    #       thus, the maximum number of left fusions is 2
    assert_cn(
        toy_gene,
        solver,
        [{"4": 2, "1": 1}],
        make_coverage(toy_gene, zip([1, 1, 1, 3, 3], [2, 2, 2, 0, 0])),
        0 + 0 + 0.5 * (base + (base * 1.5) + (base * 1.5)),
    )


def test_right_fusion(toy_gene, solver):
    base = 7.5 / len(toy_gene.unique_regions)

    # Test one fused and one normal (*1) allele (*5 is defined as 11000|11222)
    assert_cn(
        toy_gene,
        solver,
        [{"1": 1, "5": 1}],
        make_coverage(toy_gene, zip([2, 2, 1, 1, 1], [2, 2, 3, 3, 3])),
        0 + 0 + 0.5 * (base + (base * 1.25)),
    )


def test_multiplication(toy_gene, solver):
    base = 7.5 / len(toy_gene.unique_regions)

    # Test twelve copies of *1
    assert_cn(
        toy_gene,
        solver,
        [{"1": 12}],
        make_coverage(toy_gene, zip([12, 12, 12, 12, 12], [2, 2, 2, 2, 2])),
        0.5 * 12 * base,
    )
    # Test seven copies of *1 and one fused *5 copy
    assert_cn(
        toy_gene,
        solver,
        [{"1": 7, "5": 1}],
        make_coverage(toy_gene, zip([8, 8, 7, 7, 7], [2, 2, 3, 3, 3])),
        0.5 * (7 * base + (base * 1.25)),
    )



# 786
# Aldy source: lpinterface.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import Optional, Dict, Tuple, Iterable, Callable
import importlib
import collections

from .common import log, sorted_tuple, SOLUTION_PRECISION, AldyException


SOLVER_PRECISON = 1e-5
"""Default solver precision"""


def escape_name(s: str, d: Optional[dict] = None) -> str:
    """
    Escape variable names to conform with the various solver requirements.
    """
    s = s.replace(".", "").replace("-", "m").replace("#", "__").replace(">", "")[:200]
    # Ensure that all names are unique
    if d is not None:
        d[s] += 1
        if d[s] > 1:
            return s + f"_{d[s]}"
    return s


class NoSolutionsError(Exception):
    """Raised if a model is infeasible."""

    pass


class Gurobi:  # pragma: no cover
    """Wrapper around Gurobi's Python interface (:py:mod:`gurobipy`)."""

    def __init__(self, name, prev_model=None):
        self.gurobipy = importlib.import_module("gurobipy")
        self.names = collections.defaultdict(int)

        self.env = self.gurobipy.Env(empty=True)
        self.env.setParam("OutputFlag", 0)
        self.env.start()
        self.INF = self.gurobipy.GRB.INFINITY
        self.GUROBI_STATUS = {
            getattr(self.gurobipy.GRB.status, v): v
            for v in dir(self.gurobipy.GRB.status)
            if v[:2] != "__"
        }
        if prev_model:
            self.model = prev_model
        else:
            self.model = self.gurobipy.Model(name, env=self.env)
            self.model.reset()

    def addConstr(self, *args, **kwargs):
        """Add a constraint to the model."""
        if "name" in kwargs:
            kwargs["name"] = escape_name(kwargs["name"], self.names)
        c = self.model.addConstr(*args, **kwargs)
        return c

    def addVar(self, *args, **kwargs):
        """
        Add a variable to the model.

        `vtype` is the variable type:

            - `B` for binary variable
            - `I` for integer variable
            - `C` or nothing for continuous variable.
        """
        if "vtype" in kwargs and kwargs["vtype"] == "B":
            kwargs["vtype"] = self.gurobipy.GRB.BINARY
        elif "vtype" in kwargs and kwargs["vtype"] == "I":
            kwargs["vtype"] = self.gurobipy.GRB.INTEGER
        if "name" in kwargs:
            kwargs["name"] = escape_name(kwargs["name"], self.names)
        update = True
        if "update" in kwargs:
            update = kwargs["update"]
            del kwargs["update"]
        v = self.model.addVar(*args, **kwargs)
        if update:
            self.update()
        return v

    def setObjective(self, objective, method: str = "min"):
        """Set the model objective."""
        self.objective = objective
        self.model.setObjective(
            self.objective,
            (
                self.gurobipy.GRB.MINIMIZE
                if method == "min"
                else self.gurobipy.GRB.MAXIMIZE
            ),
        )
        self.update()

    def quicksum(self, expr: Iterable):
        """
        Perform a quick summation of the iterable `expr`.
        Much faster than Python's `sum` on large iterables.
        """
        return self.gurobipy.quicksum(expr)

    def update(self) -> None:
        """
        Update the model.
        Avoid calling it too much as it slows down the model construction.
        """
        self.model.update()

    def varName(self, var):
        """Return a variable name."""
        return var.varName

    def abssum(self, vars: Iterable, coeffs: Optional[Dict[str, float]] = None):
        r"""
        Return the absolute sum of `vars`: e.g.
           :math:`\sum_i |c_i x_i|` for the set :math:`{x_1,...}`.
        where :math:`c_i` is defined in the `coeffs` dictionary.

        Key of the `coeffs` dictionary stands for the name of the variable
        (should be accessible via `varName` call; 1 if not defined).
        """
        vv = []
        for i, v in enumerate(vars):
            name = self.varName(v)
            coeff = 1 if coeffs is None or name not in coeffs else coeffs[name]
            absvar = self.addVar(lb=0, update=False, name=f"ABS_{name}")
            vv.append(coeff * absvar)
            self.addConstr(absvar + v >= 0, name=f"CABSL_{i}")
            self.addConstr(absvar - v >= 0, name=f"CABSR_{i}")
        self.update()
        return self.quicksum(vv)

    def prod(self, res, terms):
        r"""
        Ensure that :math:`res = \prod terms`
        (where `terms` is a sequence of binary variables)
        by adding the appropriate linear constraints.
        Returns `res`.
        """
        for v in terms:
            self.addConstr(res <= v, name="PROD")
        self.addConstr(res >= self.quicksum(terms) - (len(terms) - 1), name="PROD")
        return res

    def solve(self, init: Optional[Callable] = None) -> Tuple[str, float]:
        """
        Solve the model. Assumes that the objective is set.

        Additional parameters of the solver can be set via `init` function that takes
        the model instance as the sole argument.

        :returns: Status of the solution and the objective value.
        :raise: :py:class:`NoSolutionsError` if the model is infeasible.
        """

        self.model.params.outputFlag = 0
        self.model.params.logFile = ""
        if init is not None:
            init(self.model)
        self.model.optimize()

        status = self.GUROBI_STATUS[self.model.status]
        if self.model.status == self.gurobipy.GRB.INFEASIBLE:
            raise NoSolutionsError(status)
        return status.lower(), self.model.objVal

    def getValue(self, var):
        """
        Get the value of the solved variable.
        Automatically adjusts the return type based on the variable type.
        """
        if hasattr(var, "vtype") and var.vtype == self.gurobipy.GRB.BINARY:
            return round(var.x) > 0
        if hasattr(var, "vtype") and var.vtype == self.gurobipy.GRB.INTEGER:
            return int(round(var.x))
        elif hasattr(var, "x"):
            return var.x
        else:
            return var.getValue()

    def dump(self, file):
        """Dump the model description (in LP format) to a file."""
        self.model.write(file)

    def variables(self):
        """Return the list of model variables."""
        return self.model.getVars()

    def is_binary(self, v):
        """Check if the variable is binary."""
        return v.vtype == self.gurobipy.GRB.BINARY

    def solutions(
        self,
        gap: float = 0,
        best_obj: Optional[float] = None,
        limit=None,
        iteration=0,
        init: Optional[Callable] = None,
    ):
        """
        Solve the model and returns the list of all optimal solutions.
        Assumes that the objective is set.
        Any solution whose score is less than (1 + `gap`) times
        the optimal solution score will be included.

        A solution is defined as a dictionary of set binary variables within
        the solution that are accessed
        by their name.

        Additional parameters of the solver can be set via `init` function that takes
        the model instance as the sole argument.

        This is a generic version that supports any solver.

        :yields: Status of the solution, the objective value and the solution itself.
        """

        try:
            status, obj = self.solve(init)
            best_obj = obj if best_obj is None else best_obj
            if status != "optimal":
                return
            ub = (1 + gap) * best_obj
            if abs(obj - ub) >= SOLVER_PRECISON and obj > ub:
                return

            vv = {
                self.varName(v): v
                for v in self.variables()
                if self.is_binary(v) and self.getValue(v) == 1
            }
            yield status, obj, sorted_tuple(set(vv.keys()))

            if not limit or iteration + 1 < limit:
                self.addConstr(self.quicksum(vv.values()) <= len(vv) - 1)
                yield from self.solutions(gap, best_obj, limit, iteration + 1, init)
        except NoSolutionsError:
            return


class CBC(Gurobi):
    """
    Wrapper around CBC's Python interface (Google's ortools).
    """

    def __init__(self, name):
        self.ortools = importlib.import_module("ortools.linear_solver.pywraplp")
        self.model = self.ortools.Solver(
            name, self.ortools.Solver.CBC_MIXED_INTEGER_PROGRAMMING
        )
        self.INF = self.model.infinity()
        self.STATUS = collections.defaultdict(
            lambda: "UNKNOWN",
            {
                self.ortools.Solver.OPTIMAL: "OPTIMAL",
                self.ortools.Solver.FEASIBLE: "FEASIBLE",
                self.ortools.Solver.INFEASIBLE: "INFEASIBLE",
                self.ortools.Solver.UNBOUNDED: "UNBOUNDED",
                self.ortools.Solver.ABNORMAL: "ABNORMAL",
                self.ortools.Solver.NOT_SOLVED: "NOT_SOLVED",
            },
        )
        self.names = collections.defaultdict(int)

    def update(self):
        pass

    def addConstr(self, *args, **kwargs):
        if "name" in kwargs:
            kwargs["name"] = escape_name(kwargs["name"], self.names)
        return self.model.Add(*args, **kwargs)

    def addVar(self, *_, **kwargs):
        name = escape_name(kwargs.get("name", ""), self.names)
        lb = kwargs.get("lb", 0)
        ub = kwargs.get("ub", self.INF)
        if "vtype" in kwargs and kwargs["vtype"] == "B":
            v = self.model.BoolVar(name)
        elif "vtype" in kwargs and kwargs["vtype"] == "I":
            v = self.model.IntVar(lb, ub, name)
        else:
            v = self.model.NumVar(lb, ub, name)
        return v

    def setObjective(self, objective, method: str = "min"):
        self.objective = objective
        if method == "min":
            self.model.Minimize(self.objective)
        else:
            self.model.Maximize(self.objective)

    def quicksum(self, expr):
        return self.model.Sum(expr)

    def solve(self, init: Optional[Callable] = None) -> Tuple[str, float]:
        if init is not None:
            init(self.model)
        status = self.model.Solve()

        if status == self.ortools.Solver.INFEASIBLE:
            raise NoSolutionsError(status)
        if not self.model.VerifySolution(SOLVER_PRECISON, True):
            raise NoSolutionsError(status)
        return self.STATUS[status].lower(), self.model.Objective().Value()

    def varName(self, var):
        return var.name()

    def getValue(self, var):
        x = var.solution_value()
        if hasattr(var, "integer") and var.integer():
            x = int(round(x))
            if (
                abs(var.lb()) < SOLUTION_PRECISION
                and abs(1 - var.ub()) < SOLUTION_PRECISION
            ):
                return x > 0
            else:
                return x
        else:
            return x

    def dump(self, file):
        content = self.model.ExportModelAsLpFormat(False)
        with open(file, "w") as f:
            f.write(content)

    def variables(self):
        return self.model.variables()

    def is_binary(self, v):
        return isinstance(self.getValue(v), bool)


def model(name: str, solver: str):
    """
    Create an ILP solver instance for a model named `name`.
    If `solver` is `'any'`, this function will try to use
    Gurobi, and will fall back on CBC if Gurobi is missing.

    :raise: :py:class:`Exception` if no solver is found.
    """

    def test_gurobi(name):  # pragma: no cover
        """Test if Gurobi is present. Requires Gurobi 7+."""
        try:
            model = Gurobi(name)
            log.trace("[lp] solver= gurobi")
        except ImportError:
            model = None
        if model and not model.model:
            model = None
        return model

    def test_cbc(name):
        """Test if OR-Tools are present. Requires Google's `ortools`."""
        try:
            model = CBC(name)
            log.trace("[lp] solver= cbc")
        except ImportError:
            model = None
        return model

    if solver == "any":
        model = test_cbc(name)
        if model is None:
            model = test_gurobi(name)
        if model is None:
            raise AldyException(
                "No ILP solver found. Aldy cannot operate without an ILP solver. "
                + "Please install Gurobi or Google OR Tools."
            )
        return model
    else:
        fname = "test_" + solver
        if fname in locals():
            m = locals()[fname](name)
            if not m:
                raise AldyException(
                    "ILP solver {} cannot be initialized".format(solver)
                )
            return m
        else:
            raise AldyException("ILP solver {} is not supported".format(solver))



#cython: profile=False

import re
import random
cimport cython

from cpython cimport array
import array

from  functools import partial
from difflib import get_close_matches, SequenceMatcher


from .variant cimport Variant
from .local_reference cimport UnsplicedLocalReference

from .consensus import consensus_refseq
from .gappedaln import find_by_normalization

from .utilities cimport (
    split,
    locate_indels,
    get_spliced_subreads,
    count_lowqual_non_ref_bases,
)
from .utilities import (
    split_cigar,
    most_common,
    to_flat_list,
    get_mapped_subreads,
    get_local_reference,
    make_insertion_first,
)
from .localn import (
    align,
    make_aligner,
    is_worth_realn,
    findall_indels,
    findall_mismatches,
)

from pysam.libcfaidx cimport FastaFile
from pysam.libcalignedsegment cimport AlignedSegment
from pysam.libcalignmentfile cimport AlignmentFile

random.seed(123)

cigar_ptrn = re.compile(r"[0-9]+[MIDNSHPX=]")


cdef tuple make_pileup(
    Variant target,
    AlignmentFile bam,
    UnsplicedLocalReference unspl_loc_ref,
    bint exclude_duplicates,
    int window,
    int downsamplethresh,
    int basequalthresh,
):
    cdef str chrom
    cdef int pos
    cdef FastaFile reference
    cdef AlignedSegment seg
    cdef dict read

    chrom, pos, reference = target.chrom, target.pos, target.reference
    rpos = max(v.pos for v in target.generate_equivalents())

    ref_len = reference.get_reference_length(chrom)

    chroms = bam.references
    if chrom not in chroms:
        if chrom.startswith("chr"):
            _chrom = chrom.replace("chr","")
        else:
            _chrom = "chr" + chrom
    else:
        _chrom = chrom

    pileup = fetch_reads(_chrom, pos, bam, ref_len, window, exclude_duplicates)
    call_back = "all" if exclude_duplicates else "nofilter"
    orig_depth = bam.count(_chrom, pos - 1, pos, read_callback=call_back)
    orig_read_num = len(pileup)

    # downsampling
    if orig_depth > downsamplethresh:
        random.seed(123)
        n_sample = int(orig_read_num * (downsamplethresh / orig_depth))

        ### lower bounded by downsample thresh / 2
        ### to prevent over-downsampling
        ### example: orig_depth = 1197119 (exclude_duplicates=False)
        ###          orig_read_num = 1962
        ###          downsamplethresh = 1000
        ###          n_sample -> 1
        ###          over downsampled.
        if n_sample >= downsamplethresh / 2 > 0:
            pileup = random.sample(pileup,  n_sample)
            sample_factor = orig_read_num / len(pileup)
        else:
            sample_factor = 1.0
    else:
        sample_factor = 1.0

    pileup = [
        dictize_read(seg, chrom, pos, rpos, reference, unspl_loc_ref, basequalthresh) for seg in pileup
    ]

    pileup = [read for read in pileup if not is_within_intron(read, pos, window)]

    return pileup, sample_factor


def is_within_intron(read, pos, window):
    intron = read["intron_pattern"]
    if intron == (0, 0):
        return False
    else:
        intron_start, intron_end = intron[0], intron[1]
        if intron_start < pos - window and pos + window < intron_end:
            return True
        else:
            return False


cdef list fetch_reads(str chrom, int pos, AlignmentFile bam, int ref_len, int window, bint exclude_duplicates):

    cdef AlignedSegment read

    pos = pos - 1  # convert to 0-based

    all_reads = bam.fetch(
        chrom, max(0, pos - window), min(pos + 1 + window, ref_len), until_eof=True
    )

    if exclude_duplicates:
        reads = [
            read
            for read in all_reads
            if not read.is_duplicate
            and not read.is_secondary
            and read.cigarstring
            and read.reference_start
        ]
    else:
        reads = [
            read
            for read in all_reads
            if not read.is_secondary
            and read.cigarstring
        ]

    return reads


cdef dict dictize_read(
    AlignedSegment read,
    str chrom,
    int pos,
    int rpos,
    FastaFile reference,
    UnsplicedLocalReference unspl_loc_ref,
    int basequalthresh
):

    cdef tuple ins, deln

    cigar_string = read.cigarstring
    cigar_list = cigar_ptrn.findall(cigar_string)

    # adjust read start/end if starts or ends with softclipping
    aln_start = read.reference_start + 1
    start_offset = int(cigar_list[0][:-1]) if cigar_list[0].endswith("S") else 0
    read_start = aln_start - start_offset

    aln_end = read.reference_end  # do not add 1
    if aln_end is None:
        aln_end = aln_start + sum(int(c[:-1]) for c in cigar_list if c[-1] in ("M", "N", "D", "=", "X"))

    end_offset = int(cigar_list[-1][:-1]) if cigar_list[-1].endswith("S") else 0

    read_end = aln_end + end_offset

    read_seq = read.query_sequence
    read_qual = read.query_qualities
    ref_seq = get_ref_seq(chrom, aln_start, aln_end, cigar_string, cigar_list, reference, unspl_loc_ref)

    read_dict = {
        "read": read,
        "read_seq": read_seq,
        "read_qual": read_qual,
        "ref_seq": ref_seq,
        "is_reverse": read.is_reverse,
        "read_name": read.query_name,
        "mapq": read.mapping_quality,
        "start_offset": start_offset,
        "aln_start": aln_start,
        "read_start": read_start,
        "end_offset": end_offset,
        "aln_end": aln_end,
        "read_end": read_end,
        "cigar_string": cigar_string,
        "cigar_list": cigar_list,
        "is_reference_seq": (read_seq == ref_seq),
        "I": [],
        "D": [],
    }

    # base qual check
    read_dict["low_qual_base_num"] = count_lowqual_non_ref_bases(read_seq, ref_seq, read_qual, cigar_list, basequalthresh)
    read_dict["is_end_dirty"] = is_end_dirty(read_qual, basequalthresh, pos, read_start, read_end, cigar_string)
    read_dict["is_dirty"] = sum(q <= basequalthresh for q in read_qual) / len(read_seq) > 0.15

    insertions, deletions = locate_indels(cigar_string, read_start)

    for ins in insertions:
        read_dict["I"].append(
            leftalign_indel_read(
                chrom,
                ins[0],
                ins[1],
                "I",
                cigar_string,
                read_start,
                aln_start,
                read_seq,
                ref_seq,
                read_qual,
                reference,
            )
        )

    for deln in deletions:
        read_dict["D"].append(
            leftalign_indel_read(
                chrom,
                deln[0],
                deln[1],
                "D",
                cigar_string,
                read_start,
                aln_start,
                read_seq,
                ref_seq,
                read_qual,
                reference,
            )
        )

    indels = [ins[-1] for ins in read_dict["I"]] + [deln[-1] for deln in read_dict["D"]]

    (
        is_covering,
        covering_subread,
        is_spliced,
        splice_ptrn,
        intron_ptrn,
    ) = parse_spliced_read(cigar_string, read_start, read_end, pos, rpos)

    read_dict["is_covering"] = is_covering
    read_dict["covering_subread"] = covering_subread
    read_dict["is_spliced"] = is_spliced
    read_dict["splice_pattern"] = splice_ptrn
    read_dict["intron_pattern"] = intron_ptrn

    return read_dict


cdef str get_ref_seq(
    str chrom,
    int aln_start,
    int aln_end,
    str cigar_string,
    list cigar_list,
    FastaFile reference,
    UnsplicedLocalReference unspl_loc_ref,
):
    cdef int current_pos = aln_start - 1

    if not "N" in cigar_string:
        #reference.fetch(chrom, current_pos, aln_end)
        return unspl_loc_ref.get_ref_seq(current_pos, aln_end)

    cdef str ref_seq = ""
    cdef str event, cigar
    cdef int event_len

    for cigar in cigar_list:
        event, event_len = cigar[-1], int(cigar[:-1])
        if event == "M" or event == "D":
            ref_seq += reference.fetch(chrom, current_pos, current_pos + event_len)
            current_pos += event_len
        elif event in ("I", "S", "H", "P"):
            pass
        else:
            current_pos += event_len

    return ref_seq


cdef tuple leftalign_indel_read(
    str chrom,
    int pos,
    int indel_len,
    str indel_type,
    str cigar_string,
    int read_start,
    int aln_start,
    str read_seq,
    str ref_seq,
    object read_qual,
    FastaFile reference,
):
    lt_flank, rt_flank = split(
        read_seq, cigar_string, pos, read_start, is_for_ref=False, reverse=False
    )
    lt_ref, rt_ref = split(
        ref_seq, cigar_string, pos, aln_start, is_for_ref=True, reverse=False
    )
    lt_qual, rt_qual = split(
        read_qual, cigar_string, pos, read_start, is_for_ref=False, reverse=False
    )

    padding_base = reference.fetch(chrom, pos - 1, pos) if "N" in cigar_string or not lt_ref else lt_ref[-1]
    if indel_type == "I":
        indel_seq = rt_flank[:indel_len]
        rt_flank = rt_flank[indel_len:]
        rt_qual = rt_qual[indel_len:]
        var = Variant(chrom, pos, padding_base, padding_base + indel_seq, reference, skip_validation=True)
    else:
        indel_seq = rt_ref[:indel_len]
        rt_ref = rt_ref[indel_len:]
        var = Variant(chrom, pos, padding_base + indel_seq, padding_base, reference, skip_validation=True)

    return pos, lt_flank, indel_seq, rt_flank, lt_ref, rt_ref, lt_qual, rt_qual, var


cdef bint is_end_dirty(array.array read_qual, int basequalthresh, int pos, int read_start, int read_end, str cigar_string):
    cdef int dist_to_left_end = pos - read_start
    cdef int dist_to_right_end = read_end - pos
    cdef bint is_lefty

    if dist_to_left_end < 0:
        is_lefty = True
    elif dist_to_right_end < 0:
        is_lefty = False
    else:
        is_lefty = (dist_to_left_end <= dist_to_right_end)

    if cigar_string.count("N") > 1:
        return False
    else:
        if is_lefty:
            return min(read_qual[: 3]) < basequalthresh
        else:
            return min(read_qual[-3 :]) < basequalthresh


cdef str leftalign_cigar(str cigarstring, Variant target, int read_start):
    target.normalize(inplace=True)

    pos = target.pos

    lt_cigar_lst, rt_cigar_lst = split_cigar(cigarstring, pos, read_start)

    if len(rt_cigar_lst) < 3:
        return cigarstring

    tmp0, tmp1, tmp2 = rt_cigar_lst[0], rt_cigar_lst[1], rt_cigar_lst[2]
    if "M" in tmp0 and "M" in tmp2:
        tmp0, tmp2 = int(tmp0[:-1]), int(tmp2[:-1])
    else:
        return cigarstring

    new_cigar = tmp1 + str(tmp0 + tmp2) + "M" + "".join(rt_cigar_lst[3:])

    return "".join(lt_cigar_lst) + new_cigar


cdef tuple parse_spliced_read(str cigar_string, int read_start, int read_end, int pos, int rpos):
    spliced_subreads = get_spliced_subreads(cigar_string, read_start, read_end)


    is_covering = False
    covering_subread = None
    for subread in spliced_subreads:
        if subread[0] <= pos <= subread[1]:
            is_covering = True
            covering_subread = tuple(subread)
        elif subread[0] <= rpos <= subread[1]:
            is_covering = True
            covering_subread = tuple(subread)
            pos = rpos


    intron_ptrn = (0, 0)
    if len(spliced_subreads) > 1:
        is_spliced = True

        lt_ptrn, rt_ptrn = "", ""
        positions = to_flat_list(spliced_subreads)
        positions = positions[1:-1]

        i = 0
        while i < len(positions):
            start = positions[i] + 1
            end = positions[i + 1] - 1

            if end < pos:
                if not lt_ptrn:
                    lt_ptrn += str(start) + "-" + str(end)
                else:
                    lt_ptrn += ":" + str(start) + "-" + str(end)

            elif pos < start - 1:
                if not rt_ptrn:
                    rt_ptrn += str(start) + "-" + str(end)
                else:
                    rt_ptrn += ":" + str(start) + "-" + str(end)

            # overhang reads
            if start - 4 <= pos <= end:
                intron_ptrn = (start, end)

            i += 2

        splice_ptrn = (lt_ptrn, rt_ptrn)
    else:
        is_spliced = False
        splice_ptrn = ("", "")

    return is_covering, covering_subread, is_spliced, splice_ptrn, intron_ptrn


def check_overhangs(pileup, splice_rate=0.2):

    intron_ptrns = [read["intron_pattern"] for read in pileup if is_junctional(read)]
    introns = [ptrn for ptrn in intron_ptrns if ptrn != (0, 0)]
    if not introns:
        return None
    else:
        intron = most_common(introns)
        if intron_ptrns.count(intron) / len(intron_ptrns) < splice_rate:
            return None

    intron_start, intron_end = intron[0], intron[1]
    overhangs = [read for read in pileup if is_overhang(read, intron_start, intron_end)]
    if overhangs:
        return intron, overhangs
    else:
        return None


def is_junctional(read):
    if read["intron_pattern"] == (0, 0):
        return read["is_covering"]
    else:
        return True


def is_overhang(read, intron_start, intron_end):
    covering_subread = read["covering_subread"]
    if not covering_subread:
        return False

    lt_read_lim = max(covering_subread[0], read["aln_start"])
    rt_read_lim = min(covering_subread[1], read["aln_end"])

    if lt_read_lim < intron_start and rt_read_lim < intron_end:
        return True
    elif intron_start < lt_read_lim and intron_end < rt_read_lim:
        return True
    else:
        return False


def overhang_aligners(target, intron, match_score, mismatch_penalty):
    genome_ref = target.reference.fetch(
        target.chrom, target.pos - 100, target.pos + 100
    )
    genome_aligner = make_aligner(genome_ref, match_score, mismatch_penalty)

    lt_exon_end, rt_exon_start = intron[0] - 1, intron[1]

    junction_ref = target.reference.fetch(
        target.chrom, lt_exon_end - 100, lt_exon_end
    ) + target.reference.fetch(target.chrom, rt_exon_start, rt_exon_start + 100)

    junction_aligner = make_aligner(junction_ref, match_score, mismatch_penalty)

    return genome_aligner, junction_aligner


def filter_spurious_overhangs(
    target,
    intron,
    overhangs,
    match_score,
    mismatch_penalty,
    gap_open_penalty,
    gap_extension_penalty,
):

    genome_aligner, junctional_aligner = overhang_aligners(
        target, intron, match_score, mismatch_penalty
    )
    non_spurious_overhangs = [
        read
        for read in overhangs
        if not read["is_reference_seq"]
        and is_non_spurious_overhang(
            read,
            target,
            intron,
            genome_aligner,
            junctional_aligner,
            match_score,
            mismatch_penalty,
            gap_open_penalty,
            gap_extension_penalty,
        )
    ]

    return non_spurious_overhangs


def is_non_spurious_overhang(
    read,
    target,
    intron,
    genome_aligner,
    junction_aligner,
    match_score,
    mismatch_penalty,
    gap_open_penalty,
    gap_extension_penalty,
):
    read_seq = read["read_seq"]

    genome_aln = align(
        genome_aligner, read_seq, gap_open_penalty, gap_extension_penalty
    )
    junction_aln = align(
        junction_aligner, read_seq, gap_open_penalty, gap_extension_penalty
    )

    genome_score = genome_aln.optimal_score
    junction_score = junction_aln.optimal_score
    if genome_score <= junction_score:
        return False

    genome_cigar = make_insertion_first(genome_aln.CIGAR)
    gap_cnt = genome_cigar.count("I") + genome_cigar.count("D")

    if gap_cnt > 3:
        return False
    elif 1 < gap_cnt <= 3:
        if genome_score / junction_score < 1.2 or genome_score < match_score * 50:
            return False
    elif gap_cnt == 0:
        aln_len = genome_aln.read_end - genome_aln.read_start + 1
        if aln_len / len(read_seq) > 0.98:
            return False

    lt_exon_end, rt_exon_start = intron[0] - 1, intron[1]
    indels_within_intron = [
        lt_exon_end < var[-1].pos < rt_exon_start for var in read["D"] and read["I"]
    ]

    if indels_within_intron:
        return True

    read = findall_mismatches(read)
    return is_worth_realn(read, target)


def retarget(
    target,
    pileup,
    window,
    mapq4retarget,
    within,
    retargetcutoff,
    match_score,
    mismatch_penalty,
    gap_open_penalty,
    gap_extension_penalty,
    unspl_loc_ref,
    require_exact_for_shiftable,
):
    target_seq, target_type, target_pos  = target.indel_seq, target.variant_type, target.pos


    if target.is_ins:
        non_refs = [
            read_dict
            for read_dict in pileup
            if not read_dict["is_reference_seq"]
            and read_dict["is_covering"]
            and read_dict["mapq"] > mapq4retarget
        ]
    else:
        non_refs = [
            read_dict
            for read_dict in pileup
            if not read_dict["is_reference_seq"]
            and read_dict["mapq"] > mapq4retarget
        ]

    if not non_refs:
        return None

    # cutoff auto_adjustment
    if len(target.indel_seq) < 3:
        cutoff = 1.0
    else:
        cutoff = retargetcutoff

    #best_matches = None
    # check by realn if no gapped alignments or no good gapped alignmetns
    #if not candidates or not best_matches:
    #    is_gapped_aln = False

    tmp_non_refs = non_refs.copy()

    non_refs = [
        read_dict
        for read_dict
        in non_refs
        if read_dict["low_qual_base_num"] < 6
        and not read_dict["is_dirty"]
        and not read_dict["is_end_dirty"]
        and read_dict.get("is_worth_realn", True)
    ]

    if not non_refs:
        non_refs = [read_dict for read_dict in tmp_non_refs if not read_dict["is_dirty"]]

    ref_starts, ref_alns, ref_seqs, aligners = [], [], [], []
    for read in non_refs:
        ref_seq, lt_len = get_local_reference(target, [read], window, unspl_loc_ref)
        ref_seqs.append(ref_seq)

        aligner = make_aligner(ref_seq, match_score, mismatch_penalty)
        aligners.append(aligner)

        ref_alns.append(align(aligner, read["read_seq"], gap_open_penalty, gap_extension_penalty))
        ref_starts.append(target.pos + 1 - lt_len)

    complex_flags = []
    candidates, candidate_reads, candidate_ref_seqs, candidate_ref_starts, candidate_aligners = [], [], [], [], []
    for read, aln, ref_seq, ref_start, aligner in zip(non_refs, ref_alns, ref_seqs, ref_starts, aligners):

        #this could happen
        if not aln.CIGAR:
            continue

        genome_aln_pos = ref_start + aln.reference_start

        aligned_read_len = (aln.read_end - aln.read_start)
        window_len = window * 6
        aligned_frac = aligned_read_len / min(len(read["read_seq"]), window_len)

        gap_cnt = aln.CIGAR.count("I") + aln.CIGAR.count("D")

        if 0 < gap_cnt < 6 and aligned_frac > 0.7:
            indels = findall_indels(aln, genome_aln_pos, ref_seq, read["read_seq"])
            positions = [d["pos"] for d in indels]
            complex_positions = set([p for p in positions if positions.count(p) == 2])

            target_type_indels = [
                indel for indel in indels if indel["indel_type"] == target_type
            ]

            # check for possible complex case
            if complex_positions:
                complex_flags.append(1)
            else:
                pass

            for indel in target_type_indels:
                if indel["pos"] in complex_positions:
                    complex_del = [jndel for jndel in indels if jndel["pos"] == indel["pos"] and jndel["indel_type"] == "D"][0]
                    complex_ins = [jndel for jndel in indels if jndel["pos"] == indel["pos"] and jndel["indel_type"] == "I"][0]
                    ref = complex_del["lt_ref"][-1] + complex_del["del_seq"]
                    alt = complex_ins["lt_ref"][-1] + complex_ins["indel_seq"]
                else:
                    if target_type == "I":
                        ref = indel["lt_ref"][-1]
                        alt = ref + indel["indel_seq"]
                    else:
                        alt = indel["lt_ref"][-1]
                        ref = alt + indel["del_seq"]

                var = Variant(target.chrom, indel["pos"], ref, alt, target.reference, skip_validation=True)

                # non-target indel found in read end -> do not consider
                read_end_thresh = max(len(read["read_seq"]) / 30, 3)
                if var.pos - read["read_start"] <= read_end_thresh or read["read_end"] - var.pos <= read_end_thresh:
                    if var == target or (complex_positions and var.pos not in complex_positions):
                        candidates.append(var)
                        candidate_reads.append(read)
                        candidate_ref_seqs.append(ref_seq)
                        candidate_ref_starts.append(ref_start)
                        candidate_aligners.append(aligner)
                else:
                    candidates.append(var)
                    candidate_reads.append(read)
                    candidate_ref_seqs.append(ref_seq)
                    candidate_ref_starts.append(ref_start)
                    candidate_aligners.append(aligner)

    if not candidates:
        # long reference may align ins as del
        if target.is_ins and window > 3:
            window = int(window/3)
            return retarget(
                        target,
                        pileup,
                        window,
                        mapq4retarget,
                        within,
                        retargetcutoff,
                        match_score,
                        mismatch_penalty,
                        gap_open_penalty,
                        gap_extension_penalty,
                        unspl_loc_ref,
                        require_exact_for_shiftable,
                   )
        else:
            return None
    elif len(target.indel_seq) <= 3:
        if not sum(complex_flags) and target not in candidates:
            return None

    u_candidates = to_flat_list([var._generate_equivalents_private() for var in set(candidates)])
    u_candidates.sort(key=lambda x: abs(x.pos - target.pos))
    candidate_seqs = [var._get_indel_seq(how=target_type) for var in u_candidates]

    best_match = get_close_matches(
        target.indel_seq, candidate_seqs, n=1, cutoff=cutoff
    )

    if best_match:
        best_seq = best_match[0]
        match_score = SequenceMatcher(None, target.indel_seq, best_seq).ratio()

        idx = candidate_seqs.index(best_seq)
        hit = u_candidates[idx]

        # require exact match for shiftable indels
        if require_exact_for_shiftable:
            if len(hit.generate_equivalents()) > 1 or len(target.generate_equivalents()) > 1:
                if hit != target:
                    return None

        if abs(target.pos - hit.pos) < within:

            try:
                idx2 = candidates.index(hit)  #get original representation, do not normalize here.
            except:
                hit.pos = hit.pos - len(hit.ref)
                idx2 = candidates.index(hit)

            candidate = candidates[idx2]
            idx = [i for i, var in enumerate(candidates) if var == candidate]

            # check if the candidate can be del/ins component of a complex indel
            if candidate.is_non_complex_indel():
                complex_candidates = [var for var in set(candidates) if not var.is_non_complex_indel()]
                if complex_candidates:
                    for cplx_candidate in complex_candidates:
                        reduced = cplx_candidate._reduce_complex_indel(to=target_type)
                        if candidate == reduced:
                            idx = [i for i, var in enumerate(candidates) if var == cplx_candidate]
                            candidate = reduced
                            break
            else:
                candidate = candidate._reduce_complex_indel(to=target_type)

            candidate_reads = [candidate_reads[i] for i in idx]
            candidate_ref_seqs = [candidate_ref_seqs[i] for i in idx]
            candidate_ref_starts = [candidate_ref_starts[i] for i in idx]
            candidate_aligners = [candidate_aligners[i] for i in idx]

            return candidate, candidate_reads, match_score, candidate_ref_seqs, candidate_ref_starts, candidate_aligners
        else:
            return None
    else:
        if target.is_ins and window > 3:
            window = int(window/3)
            return retarget(
                        target,
                        pileup,
                        window,
                        mapq4retarget,
                        within,
                        retargetcutoff,
                        match_score,
                        mismatch_penalty,
                        gap_open_penalty,
                        gap_extension_penalty,
                        unspl_loc_ref,
                        require_exact_for_shiftable
                    )
        else:
            return None


def update_read_info(
    read,
    candidate,
    is_gapped_aln=True,
    gap_open_penalty=3,
    gap_extension_penalty=1,
    aligner=None,
    ref_seq=None,
    ref_start=None,
):
    if is_gapped_aln:
        parsed = leftalign_indel_read(
            candidate.chrom,
            candidate.pos,
            len(candidate.indel_seq),
            candidate.variant_type,
            read["cigar_string"],
            read["read_start"],
            read["aln_start"],
            read["read_seq"],
            read["ref_seq"],
            read["read_qual"],
            candidate.reference,
        )
        read["lt_flank"] = parsed[1]
        read["indel_seq"] = parsed[2] if candidate.is_ins else ""
        read["rt_flank"] = parsed[3]
        read["lt_ref"] = parsed[4]
        read["rt_ref"] = parsed[5]
        read["lt_qual"] = parsed[6]
        read["rt_qual"] = parsed[7]

        read["lt_cigar"], read["rt_cigar"] = split_cigar(
            read["cigar_string"], candidate.pos, read["read_start"]
        )

        read["is_target"] = True
    else:
        aln = align(aligner, read["read_seq"], gap_open_penalty, gap_extension_penalty)
        genome_aln_pos = ref_start + aln.reference_start

        indels = findall_indels(
            aln, genome_aln_pos, ref_seq, read["read_seq"], basequals=read["read_qual"]
        )

        is_found= False
        for indel in indels:
            if not indel.get("del_seq", False):
                ref = indel["lt_ref"][-1]
                alt = ref + indel["indel_seq"]
            else:
                alt = indel["lt_ref"][-1]
                ref = alt + indel["del_seq"]

            obj = Variant(candidate.chrom, indel["pos"], ref, alt, candidate.reference, skip_validation=True)
            if candidate == obj:
                is_found = True
                indel_pos_in_this_read = indel["pos"]
                break

        if not is_found:
            read["cigar_updated"] = False
            return read

        read["lt_flank"] = indel["lt_flank"]
        read["indel_seq"] = candidate.indel_seq if candidate.is_ins else ""
        read["rt_flank"] = indel["rt_flank"]
        read["lt_qual"] = indel["lt_qual"]
        read["rt_qual"] = indel["rt_qual"]

        realn_lt_cigar, realn_rt_cigar = split_cigar(
            make_insertion_first(aln.CIGAR), indel["pos"], genome_aln_pos
        )

        read["lt_ref"] = trim_ref_flank(indel["lt_ref"], realn_lt_cigar, left=True)
        read["rt_ref"] = trim_ref_flank(indel["rt_ref"], realn_rt_cigar, left=False)

        read["lt_cigar"] = update_cigar(
            read["cigar_string"],
            realn_lt_cigar,
            read["read_start"],
            read["splice_pattern"],
            indel["lt_clipped"],
            left=True,
        )
        read["rt_cigar"] = update_cigar(
            read["cigar_string"],
            realn_rt_cigar,
            candidate.pos,
            read["splice_pattern"],
            indel["rt_clipped"],
            left=False,
        )

        read["cigar_list"] = read["lt_cigar"] + read["rt_cigar"]
        read["cigar_string"] = "".join(read["cigar_list"])
        read["cigar_updated"] = True

        update_read_positions(read, indel_pos_in_this_read)

        read["is_target"] = True

    return read


def trim_ref_flank(ref_flank, flank_cigar, left):

    cum = sum([int(c[:-1]) for c in flank_cigar if c[-1] != "I"])
    if left:
        ref_flank = ref_flank[-cum:]
    else:
        ref_flank = ref_flank[:cum]

    return ref_flank


def update_cigar(
    orig_cigar_string, realn_cigar, start_pos, splice_prtn, clipped_bases, left
):

    splice_ptrn = splice_prtn[0] if left else splice_prtn[1]

    if splice_ptrn:
        spl_spans = [numeric_span(spl_span) for spl_span in splice_ptrn.split(":")]
    else:
        spl_spans = []

    clip_len = len(clipped_bases)

    if left:
        new_cigar = [str(clip_len) + "S"] if clip_len else []
        current_pos = start_pos + clip_len
    else:
        new_cigar = []
        target_event = realn_cigar[0]
        target_type, target_len = target_event[-1], int(target_event[:-1])
        current_pos = (
            start_pos + 1 if target_type == "I" else start_pos + target_len + 1
        )
        trailing_clip = [str(clip_len) + "S"] if clip_len else []
        realn_cigar = realn_cigar[1:]

    for c in realn_cigar:
        event, event_len = c[-1], int(c[:-1])
        if event == "M":
            if spl_spans:
                last = len(spl_spans) - 1
                tmp = spl_spans.copy()
                for i, span in enumerate(tmp):
                    n = span[1] - span[0] + 1
                    if span[0] <= current_pos + event_len:
                        if i != last:
                            m = span[0] - current_pos

                            if m:
                                new_cigar += [str(m) + "M", str(n) + "N"]
                            else:
                                new_cigar += [str(n) + "N"]

                            current_pos += m + n
                            event_len -= m
                        else:
                            m1 = span[0] - current_pos
                            m2 = event_len - m1

                            if m2:
                                if m1:
                                    new_cigar += [str(m1) + "M", str(n) + "N", str(m2) + "M"]
                                else:
                                    new_cigar += [str(n) + "N", str(m2) + "M"]
                            else:
                                new_cigar += [str(event_len) + "M", str(n) + "N"]

                            current_pos += n + event_len
                        spl_spans = spl_spans[1:]
                    else:
                        new_cigar.append(str(event_len) + "M")
                        current_pos += event_len - 1
                        # hotfix -1 should be checked! Mar 12
                        break

            else:
                new_cigar.append(str(event_len) + "M")
                current_pos += event_len

        elif event == "I":

            if spl_spans:
                span = spl_spans[0]
                spl_start, spl_end = span[0], span[1]
                n = spl_end - spl_start + 1

                if spl_start == current_pos:
                    this_event = [str(event_len) + "I", str(n) + "N"]

                    new_cigar += this_event
                    current_pos += n

                    spl_spans = spl_spans[1:]
                else:
                    new_cigar.append(str(event_len) + "I")
                    current_pos += 1
            else:
                new_cigar.append(str(event_len) + "I")
                current_pos += 1

        elif event == "D":
            new_cigar.append(str(event_len) + "D")
            current_pos += event_len

    if left:
        return new_cigar
    else:
        return [target_event] + new_cigar + trailing_clip

    #return updated_cigar


def numeric_span(spl_span):
    spl_span_lst = spl_span.split("-")
    return [int(i) for i in spl_span_lst]


def update_read_positions(read, target_pos):

    left_adjust = sum([-int(c[:-1]) if c[-1] != "I" else 0 for c in read["lt_cigar"]])
    right_adjust = sum([int(c[:-1]) if c[-1] != "I" else 0 for c in read["rt_cigar"]])

    read["read_start"] = target_pos + left_adjust + 1
    read["read_end"] = target_pos + right_adjust

    lt_most_cigar = read["lt_cigar"][0]
    read["start_offset"] = int(lt_most_cigar[:-1]) if "S" in lt_most_cigar else 0

    rt_most_cigar = read["rt_cigar"][-1]
    read["end_offset"] = int(rt_most_cigar[:-1]) if "S" in rt_most_cigar else 0

    read["aln_start"] = read["read_start"] + read["start_offset"]
    read["aln_end"] = read["read_end"] - read["end_offset"]


def update_pileup(
    pileup,
    new_target,
    window,
    match_score,
    mismatch_penalty,
    gap_open_penalty,
    gap_extension_penalty,
    basequalthresh,
    bypass_search=False
):

    rpos = max(v.pos for v in new_target.generate_equivalents())
    for read in pileup:

        (
            is_covering,
            covering_subread,
            is_spliced,
            splice_ptrn,
            intron_ptrn,
        ) = parse_spliced_read(
            read["cigar_string"], read["read_start"], read["read_end"], new_target.pos, rpos
        )

        read["is_covering"] = is_covering
        read["covering_subread"] = covering_subread
        read["is_spliced"] = is_spliced
        read["splice_pattern"] = splice_ptrn
        read["intron_pattern"] = intron_ptrn

    if bypass_search:
        return new_target, pileup
    else:
        return find_by_normalization(
            new_target,
            pileup,
            window,
            match_score,
            mismatch_penalty,
            gap_open_penalty,
            gap_extension_penalty,
            basequalthresh,
        )



from pysam.libcfaidx cimport FastaFile


cdef class UnsplicedLocalReference:
    cdef:
        str chrom, ref, alt, unspliced_local_reference
        int pos, window, local_ref_start, ref_len
        FastaFile reference
        readonly int left_len
    
